@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -40,11 +40,17 @@ int connman_inet_ifup(int index);
 int connman_inet_ifdown(int index);
 
 struct connman_device *connman_inet_create_device(int index);
-connman_bool_t connman_inet_is_mac80211(int index);
+connman_bool_t connman_inet_is_cfg80211(int index);
 
 int connman_inet_set_address(int index, struct connman_ipaddress *ipaddress);
 int connman_inet_clear_address(int index);
-int connman_inet_set_gateway(int index, struct in_addr gateway);
+int connman_inet_add_host_route_vpn(int index, const char *gateway, const char *host);
+int connman_inet_add_host_route(int index, const char *host);
+int connman_inet_del_host_route(int index, const char *host);
+int connman_inet_set_gateway_address(int index, const char *gateway);
+int connman_inet_clear_gateway_address(int index, const char *gateway);
+int connman_inet_set_gateway_interface(int index);
+int connman_inet_clear_gateway_interface(int index);
 
 #ifdef __cplusplus
 }
