@@ -170,19 +170,16 @@ static struct connman_device_driver modem_driver = {
 
 static char *get_ident(const char *path)
 {
-	char *ident, *pos;
+	char *pos;
 
 	if (*path != '/')
 		return NULL;
 
-	ident = g_strdup(path + 1);
+	pos = strrchr(path, '/');
+	if (pos == NULL)
+		return NULL;
 
-	pos = ident;
-
-	while ((pos = strchr(pos, '/')) != NULL)
-		*pos = '_';
-
-	return ident;
+	return g_strdup(pos + 1);
 }
 
 static void create_service(struct connman_network *network)
@@ -378,9 +375,59 @@ done:
 	dbus_message_unref(message);
 }
 
+static gboolean registration_changed(DBusConnection *connection,
+					DBusMessage *message, void *user_data)
+{
+	const char *path = dbus_message_get_path(message);
+	struct connman_network *network = user_data;
+	DBusMessageIter iter, value;
+	const char *key;
+
+	DBG("path %s", path);
+
+	if (dbus_message_iter_init(message, &iter) == FALSE)
+		return TRUE;
+
+	dbus_message_iter_get_basic(&iter, &key);
+
+	DBG("key %s", key);
+
+	dbus_message_iter_next(&iter);
+	dbus_message_iter_recurse(&iter, &value);
+
+	if (g_strcmp0(key, "Name") == 0 ||
+			g_strcmp0(key, "Operator") == 0) {
+		const char *name;
+
+		dbus_message_iter_get_basic(&value, &name);
+		DBG("name %s", name);
+		connman_network_set_name(network, name);
+		create_service(network);
+	} else if (g_strcmp0(key, "Strength") == 0) {
+		connman_uint8_t strength;
+
+		dbus_message_iter_get_basic(&value, &strength);
+		connman_network_set_strength(network, strength);
+		connman_network_update(network);
+	}
+
+	return TRUE;
+}
+
 static int network_probe(struct connman_network *network)
 {
 	const char *path;
+	guint reg_watch;
+
+	reg_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+						OFONO_REGISTRATION_INTERFACE,
+						PROPERTY_CHANGED,
+						registration_changed,
+						network, NULL);
+	if (reg_watch == 0)
+		return -EIO;
+
+	connman_network_set_data(network, GUINT_TO_POINTER(reg_watch));
 
 	path = connman_network_get_string(network, "Path");
 
@@ -590,7 +637,12 @@ static int network_disconnect(struct connman_network *network)
 
 static void network_remove(struct connman_network *network)
 {
+	guint reg_watch;
+
 	DBG("network %p", network);
+
+	reg_watch = GPOINTER_TO_UINT(connman_network_get_data(network));
+	g_dbus_remove_watch(connection, reg_watch);
 }
 
 static int network_setup(struct connman_network *network, const char *key)
@@ -1577,42 +1629,12 @@ static int static_network_set_connected(
 				struct connman_element *parent,
 					connman_bool_t connected)
 {
-	if (connected == TRUE) {
-		struct connman_element *element;
-
-		if (parent->ipv4.address == NULL)
-			goto failed;
-
-		if (parent->ipv4.netmask == NULL)
-			goto failed;
-
-		element = connman_element_create(NULL);
-		if (element == NULL) {
-			connman_error("Can not create connman_element");
-			return -ENOMEM;
-		}
-
-		element->type = CONNMAN_ELEMENT_TYPE_IPV4;
-		element->index = parent->index;
-
-		if (connman_element_register(element, parent) < 0) {
-			connman_element_unref(element);
-			goto failed;
-		}
-	} else
+	if (connected == FALSE)
 		cleanup_ipconfig(parent);
 
 	connman_network_set_connected(pending_network, connected);
 
 	return 0;
-
-failed:
-	connman_network_set_error(pending_network,
-		CONNMAN_NETWORK_ERROR_ASSOCIATE_FAIL);
-
-	cleanup_ipconfig(parent);
-
-	return -EINVAL;
 }
 
 static gboolean pri_context_changed(DBusConnection *connection,
