@@ -741,10 +741,51 @@ static void set_configuration(struct connman_network *network)
 					CONNMAN_SERVICE_STATE_CONFIGURATION);
 }
 
+static int set_connected_fixed(struct connman_network *network)
+{
+	struct connman_service *service;
+	struct connman_element *parent, *element;
+
+	DBG("");
+
+	service = __connman_service_lookup_from_network(network);
+
+	parent = connman_network_get_element(network);
+
+	set_configuration(network);
+
+	if (parent->ipv4.address == NULL)
+		return -EINVAL;
+
+	if (parent->ipv4.netmask == NULL)
+		return -EINVAL;
+
+	element = connman_element_create(NULL);
+	if (element == NULL) {
+		connman_error("Can not create connman_element");
+		return -ENOMEM;
+	}
+
+	element->type = CONNMAN_ELEMENT_TYPE_IPV4;
+	element->index = parent->index;
+
+	if (connman_element_register(element, parent) < 0) {
+		connman_error("Can not register connman_element");
+		return -EINVAL;
+	}
+
+	network->connecting = FALSE;
+
+	connman_network_set_associating(network, FALSE);
+
+	return 0;
+}
+
 static void set_connected_manual(struct connman_network *network)
 {
 	struct connman_service *service;
 	struct connman_ipconfig *ipconfig;
+	const char *nameserver = NULL;
 	int err;
 
 	DBG("network %p", network);
@@ -762,6 +803,11 @@ static void set_connected_manual(struct connman_network *network)
 		return;
 	}
 
+	connman_element_get_value(&network->element,
+			CONNMAN_PROPERTY_ID_IPV4_NAMESERVER, &nameserver);
+	if (nameserver != NULL)
+		__connman_service_append_nameserver(service, nameserver);
+
 	__connman_ipconfig_set_gateway(ipconfig, &network->element);
 
 	network->connecting = FALSE;
@@ -819,8 +865,14 @@ static gboolean set_connected(gpointer user_data)
 		case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
 		case CONNMAN_IPCONFIG_METHOD_OFF:
 			return FALSE;
-		case CONNMAN_IPCONFIG_METHOD_MANUAL:
 		case CONNMAN_IPCONFIG_METHOD_FIXED:
+			if (set_connected_fixed(network) < 0) {
+				connman_network_set_error(network,
+					CONNMAN_NETWORK_ERROR_ASSOCIATE_FAIL);
+				return FALSE;
+			}
+			return TRUE;
+		case CONNMAN_IPCONFIG_METHOD_MANUAL:
 			set_connected_manual(network);
 			return TRUE;
 		case CONNMAN_IPCONFIG_METHOD_DHCP:
@@ -1513,6 +1565,27 @@ void connman_network_set_data(struct connman_network *network, void *data)
 	network->driver_data = data;
 }
 
+void connman_network_update(struct connman_network *network)
+{
+	switch (network->type) {
+	case CONNMAN_NETWORK_TYPE_UNKNOWN:
+	case CONNMAN_NETWORK_TYPE_VENDOR:
+		return;
+	case CONNMAN_NETWORK_TYPE_ETHERNET:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
+	case CONNMAN_NETWORK_TYPE_CELLULAR:
+	case CONNMAN_NETWORK_TYPE_WIFI:
+	case CONNMAN_NETWORK_TYPE_WIMAX:
+		break;
+	}
+
+	if (network->group != NULL)
+		__connman_service_update_from_network(network);
+
+	return;
+}
+
 static gboolean match_driver(struct connman_network *network,
 					struct connman_network_driver *driver)
 {
