@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -23,11 +23,13 @@
 #include <config.h>
 #endif
 
+#include <stdio.h>
 #include <unistd.h>
 #include <string.h>
 #include <sys/socket.h>
 #include <arpa/inet.h>
-
+#include <netinet/ether.h>
+#include <net/if_arp.h>
 #include <linux/if.h>
 #include <linux/netlink.h>
 #include <linux/rtnetlink.h>
@@ -36,9 +38,13 @@
 
 #include "connman.h"
 
+#define print(arg...) do { } while (0)
+//#define print(arg...) connman_info(arg)
+
 struct watch_data {
 	unsigned int id;
 	int index;
+	connman_rtnl_operstate_cb_t operstate;
 	connman_rtnl_link_cb_t newlink;
 	void *user_data;
 };
@@ -46,6 +52,45 @@ struct watch_data {
 static GSList *watch_list = NULL;
 static unsigned int watch_id = 0;
 
+/**
+ * connman_rtnl_add_operstate_watch:
+ * @index: network device index
+ * @callback: callback function
+ * @user_data: callback data;
+ *
+ * Add a new RTNL watch for operation state events
+ *
+ * Returns: %0 on failure and a unique id on success
+ */
+unsigned int connman_rtnl_add_operstate_watch(int index,
+			connman_rtnl_operstate_cb_t callback, void *user_data)
+{
+	struct watch_data *watch;
+
+	watch = g_try_new0(struct watch_data, 1);
+	if (watch == NULL)
+		return 0;
+
+	watch->id = ++watch_id;
+	watch->index = index;
+
+	watch->operstate = callback;
+	watch->user_data = user_data;
+
+	watch_list = g_slist_prepend(watch_list, watch);
+
+	DBG("id %d", watch->id);
+
+	if (callback) {
+		unsigned char operstate = 0;
+
+		if (operstate > 0)
+			callback(operstate, user_data);
+	}
+
+	return watch->id;
+}
+
 /**
  * connman_rtnl_add_newlink_watch:
  * @index: network device index
@@ -75,6 +120,13 @@ unsigned int connman_rtnl_add_newlink_watch(int index,
 
 	DBG("id %d", watch->id);
 
+	if (callback) {
+		unsigned int flags = __connman_ipconfig_get_flags(index);
+
+		if (flags > 0)
+			callback(flags, 0, user_data);
+	}
+
 	return watch->id;
 }
 
@@ -104,6 +156,25 @@ void connman_rtnl_remove_watch(unsigned int id)
 	}
 }
 
+static void trigger_rtnl(int index, void *user_data)
+{
+	struct connman_rtnl *rtnl = user_data;
+
+	if (rtnl->newlink) {
+		unsigned short type = __connman_ipconfig_get_type(index);
+		unsigned int flags = __connman_ipconfig_get_flags(index);
+
+		rtnl->newlink(type, index, flags, 0);
+	}
+
+	if (rtnl->newgateway) {
+		const char *gateway = __connman_ipconfig_get_gateway(index);
+
+		if (gateway != NULL)
+			rtnl->newgateway(index, gateway);
+	}
+}
+
 static GSList *rtnl_list = NULL;
 
 static gint compare_priority(gconstpointer a, gconstpointer b)
@@ -129,6 +200,8 @@ int connman_rtnl_register(struct connman_rtnl *rtnl)
 	rtnl_list = g_slist_insert_sorted(rtnl_list, rtnl,
 							compare_priority);
 
+	__connman_ipconfig_foreach(trigger_rtnl, rtnl);
+
 	return 0;
 }
 
@@ -145,11 +218,106 @@ void connman_rtnl_unregister(struct connman_rtnl *rtnl)
 	rtnl_list = g_slist_remove(rtnl_list, rtnl);
 }
 
-static void process_newlink(unsigned short type, int index,
-					unsigned flags, unsigned change)
+static const char *operstate2str(unsigned char operstate)
 {
+	switch (operstate) {
+	case IF_OPER_UNKNOWN:
+		return "UNKNOWN";
+	case IF_OPER_NOTPRESENT:
+		return "NOT-PRESENT";
+	case IF_OPER_DOWN:
+		return "DOWN";
+	case IF_OPER_LOWERLAYERDOWN:
+		return "LOWER-LAYER-DOWN";
+	case IF_OPER_TESTING:
+		return "TESTING";
+	case IF_OPER_DORMANT:
+		return "DORMANT";
+	case IF_OPER_UP:
+		return "UP";
+	}
+
+	return "";
+}
+
+static void extract_link(struct ifinfomsg *msg, int bytes,
+				struct ether_addr *address, const char **ifname,
+				unsigned int *mtu, unsigned char *operstate,
+						struct rtnl_link_stats *stats)
+{
+	struct rtattr *attr;
+
+	for (attr = IFLA_RTA(msg); RTA_OK(attr, bytes);
+					attr = RTA_NEXT(attr, bytes)) {
+		switch (attr->rta_type) {
+		case IFLA_ADDRESS:
+			if (address != NULL)
+				memcpy(address, RTA_DATA(attr), ETH_ALEN);
+			break;
+		case IFLA_IFNAME:
+			if (ifname != NULL)
+				*ifname = RTA_DATA(attr);
+			break;
+		case IFLA_MTU:
+			if (mtu != NULL)
+				*mtu = *((unsigned int *) RTA_DATA(attr));
+			break;
+		case IFLA_STATS:
+			if (stats != NULL)
+				memcpy(stats, RTA_DATA(attr),
+					sizeof(struct rtnl_link_stats));
+			break;
+		case IFLA_OPERSTATE:
+			if (operstate != NULL)
+				*operstate = *((unsigned char *) RTA_DATA(attr));
+			break;
+		case IFLA_LINKMODE:
+			break;
+		}
+	}
+}
+
+static void process_newlink(unsigned short type, int index, unsigned flags,
+			unsigned change, struct ifinfomsg *msg, int bytes)
+{
+	struct ether_addr address = {{ 0, 0, 0, 0, 0, 0 }};
+	struct ether_addr compare = {{ 0, 0, 0, 0, 0, 0 }};
+	struct rtnl_link_stats stats;
+	unsigned char operstate = 0xff;
+	const char *ifname = NULL;
+	unsigned int mtu = 0;
+	char str[18];
 	GSList *list;
 
+	memset(&stats, 0, sizeof(stats));
+	extract_link(msg, bytes, &address, &ifname, &mtu, &operstate, &stats);
+
+	snprintf(str, 18, "%02X:%02X:%02X:%02X:%02X:%02X",
+						address.ether_addr_octet[0],
+						address.ether_addr_octet[1],
+						address.ether_addr_octet[2],
+						address.ether_addr_octet[3],
+						address.ether_addr_octet[4],
+						address.ether_addr_octet[5]);
+
+	switch (type) {
+	case ARPHRD_ETHER:
+	case ARPHRD_LOOPBACK:
+	case ARPHRD_NONE:
+		__connman_ipconfig_newlink(index, type, flags,
+							str, mtu, &stats);
+		break;
+	}
+
+	if (memcmp(&address, &compare, ETH_ALEN) != 0)
+		connman_info("%s {newlink} index %d address %s mtu %u",
+						ifname, index, str, mtu);
+
+	if (operstate != 0xff)
+		connman_info("%s {newlink} index %d operstate %u <%s>",
+						ifname, index, operstate,
+						operstate2str(operstate));
+
 	for (list = rtnl_list; list; list = list->next) {
 		struct connman_rtnl *rtnl = list->data;
 
@@ -163,122 +331,265 @@ static void process_newlink(unsigned short type, int index,
 		if (watch->index != index)
 			continue;
 
+		if (operstate != 0xff && watch->operstate)
+			watch->operstate(operstate, watch->user_data);
+
 		if (watch->newlink)
 			watch->newlink(flags, change, watch->user_data);
 	}
 }
 
-static void process_dellink(unsigned short type, int index,
-					unsigned flags, unsigned change)
+static void process_dellink(unsigned short type, int index, unsigned flags,
+			unsigned change, struct ifinfomsg *msg, int bytes)
 {
+	struct rtnl_link_stats stats;
+	unsigned char operstate = 0xff;
+	const char *ifname = NULL;
 	GSList *list;
 
+	memset(&stats, 0, sizeof(stats));
+	extract_link(msg, bytes, NULL, &ifname, NULL, &operstate, &stats);
+
+	if (operstate != 0xff)
+		connman_info("%s {dellink} index %d operstate %u <%s>",
+						ifname, index, operstate,
+						operstate2str(operstate));
+
+	for (list = watch_list; list; list = list->next) {
+		struct watch_data *watch = list->data;
+
+		if (watch->index != index)
+			continue;
+
+		if (operstate != 0xff && watch->operstate)
+			watch->operstate(operstate, watch->user_data);
+	}
+
 	for (list = rtnl_list; list; list = list->next) {
 		struct connman_rtnl *rtnl = list->data;
 
 		if (rtnl->dellink)
 			rtnl->dellink(type, index, flags, change);
 	}
+
+	switch (type) {
+	case ARPHRD_ETHER:
+	case ARPHRD_LOOPBACK:
+	case ARPHRD_NONE:
+		__connman_ipconfig_dellink(index, &stats);
+		break;
+	}
 }
 
-static char *extract_gateway(struct rtmsg *msg, int bytes, int *index)
+static void extract_addr(struct ifaddrmsg *msg, int bytes,
+						const char **label,
+						struct in_addr *local,
+						struct in_addr *address,
+						struct in_addr *broadcast)
+{
+	struct rtattr *attr;
+
+	for (attr = IFA_RTA(msg); RTA_OK(attr, bytes);
+					attr = RTA_NEXT(attr, bytes)) {
+		switch (attr->rta_type) {
+		case IFA_ADDRESS:
+			if (address != NULL)
+				*address = *((struct in_addr *) RTA_DATA(attr));
+			break;
+		case IFA_LOCAL:
+			if (local != NULL)
+				*local = *((struct in_addr *) RTA_DATA(attr));
+			break;
+		case IFA_BROADCAST:
+			if (broadcast != NULL)
+				*broadcast = *((struct in_addr *) RTA_DATA(attr));
+			break;
+		case IFA_LABEL:
+			if (label != NULL)
+				*label = RTA_DATA(attr);
+			break;
+		}
+	}
+}
+
+static void process_newaddr(unsigned char family, unsigned char prefixlen,
+				int index, struct ifaddrmsg *msg, int bytes)
+{
+	struct in_addr address = { INADDR_ANY };
+	const char *label = NULL;
+
+	if (family != AF_INET)
+		return;
+
+	extract_addr(msg, bytes, &label, &address, NULL, NULL);
+
+	__connman_ipconfig_newaddr(index, label,
+					prefixlen, inet_ntoa(address));
+}
+
+static void process_deladdr(unsigned char family, unsigned char prefixlen,
+				int index, struct ifaddrmsg *msg, int bytes)
+{
+	struct in_addr address = { INADDR_ANY };
+	const char *label = NULL;
+
+	if (family != AF_INET)
+		return;
+
+	extract_addr(msg, bytes, &label, &address, NULL, NULL);
+
+	__connman_ipconfig_deladdr(index, label,
+					prefixlen, inet_ntoa(address));
+}
+
+static void extract_route(struct rtmsg *msg, int bytes, int *index,
+						struct in_addr *dst,
+						struct in_addr *gateway)
 {
-	char *gateway = NULL;
-	struct in_addr addr;
 	struct rtattr *attr;
 
 	for (attr = RTM_RTA(msg); RTA_OK(attr, bytes);
 					attr = RTA_NEXT(attr, bytes)) {
 		switch (attr->rta_type) {
+		case RTA_DST:
+			if (dst != NULL)
+				*dst = *((struct in_addr *) RTA_DATA(attr));
+			break;
 		case RTA_GATEWAY:
-			addr = *((struct in_addr *) RTA_DATA(attr));
-			g_free(gateway);
-			gateway = g_strdup(inet_ntoa(addr));
+			if (gateway != NULL)
+				*gateway = *((struct in_addr *) RTA_DATA(attr));
 			break;
 		case RTA_OIF:
-			*index = *((int *) RTA_DATA(attr));
+			if (index != NULL)
+				*index = *((int *) RTA_DATA(attr));
 			break;
 		}
 	}
-
-	return gateway;
 }
 
-static void process_newgateway(struct rtmsg *msg, int bytes)
+static void process_newroute(unsigned char family, unsigned char scope,
+						struct rtmsg *msg, int bytes)
 {
-	int index = -1;
-	char *gateway;
 	GSList *list;
+	struct in_addr dst = { INADDR_ANY }, gateway = { INADDR_ANY };
+	char dststr[16], gatewaystr[16];
+	int index = -1;
 
-	gateway = extract_gateway(msg, bytes, &index);
-	if (gateway == NULL || index < 0)
+	if (family != AF_INET)
+		return;
+
+	extract_route(msg, bytes, &index, &dst, &gateway);
+
+	inet_ntop(family, &dst, dststr, sizeof(dststr));
+	inet_ntop(family, &gateway, gatewaystr, sizeof(gatewaystr));
+
+	__connman_ipconfig_newroute(index, scope, dststr, gatewaystr);
+
+	/* skip host specific routes */
+	if (scope != RT_SCOPE_UNIVERSE &&
+			!(scope == RT_SCOPE_LINK && dst.s_addr == INADDR_ANY))
+		return;
+
+	if (dst.s_addr != INADDR_ANY)
 		return;
 
 	for (list = rtnl_list; list; list = list->next) {
 		struct connman_rtnl *rtnl = list->data;
 
 		if (rtnl->newgateway)
-			rtnl->newgateway(index, gateway);
+			rtnl->newgateway(index, gatewaystr);
 	}
-
-	g_free(gateway);
 }
 
-static void process_delgateway(struct rtmsg *msg, int bytes)
+static void process_delroute(unsigned char family, unsigned char scope,
+						struct rtmsg *msg, int bytes)
 {
-	int index = -1;
-	char *gateway;
 	GSList *list;
+	struct in_addr dst = { INADDR_ANY }, gateway = { INADDR_ANY };
+	char dststr[16], gatewaystr[16];
+	int index = -1;
 
-	gateway = extract_gateway(msg, bytes, &index);
-	if (gateway == NULL || index < 0)
+	if (family != AF_INET)
+		return;
+
+	extract_route(msg, bytes, &index, &dst, &gateway);
+
+	inet_ntop(family, &dst, dststr, sizeof(dststr));
+	inet_ntop(family, &gateway, gatewaystr, sizeof(gatewaystr));
+
+	__connman_ipconfig_delroute(index, scope, dststr, gatewaystr);
+
+	/* skip host specific routes */
+	if (scope != RT_SCOPE_UNIVERSE &&
+			!(scope == RT_SCOPE_LINK && dst.s_addr == INADDR_ANY))
+		return;
+
+	if (dst.s_addr != INADDR_ANY)
 		return;
 
 	for (list = rtnl_list; list; list = list->next) {
 		struct connman_rtnl *rtnl = list->data;
 
 		if (rtnl->delgateway)
-			rtnl->delgateway(index, gateway);
+			rtnl->delgateway(index, gatewaystr);
 	}
-
-	g_free(gateway);
 }
 
-static inline void print_inet(struct rtattr *attr, const char *name, int family)
+static inline void print_ether(struct rtattr *attr, const char *name)
 {
-	if (family == AF_INET) {
+	int len = (int) RTA_PAYLOAD(attr);
+
+	if (len == ETH_ALEN) {
+		struct ether_addr eth;
+		memcpy(&eth, RTA_DATA(attr), ETH_ALEN);
+		print("  attr %s (len %d) %s\n", name, len, ether_ntoa(&eth));
+	} else
+		print("  attr %s (len %d)\n", name, len);
+}
+
+static inline void print_inet(struct rtattr *attr, const char *name,
+							unsigned char family)
+{
+	int len = (int) RTA_PAYLOAD(attr);
+
+	if (family == AF_INET && len == sizeof(struct in_addr)) {
 		struct in_addr addr;
 		addr = *((struct in_addr *) RTA_DATA(attr));
-		DBG("  attr %s (len %d) %s\n", name,
-				(int) RTA_PAYLOAD(attr), inet_ntoa(addr));
+		print("  attr %s (len %d) %s\n", name, len, inet_ntoa(addr));
 	} else
-		DBG("  attr %s (len %d)\n", name, (int) RTA_PAYLOAD(attr));
+		print("  attr %s (len %d)\n", name, len);
 }
 
-static inline void print_char(struct rtattr *attr, const char *name)
+static inline void print_string(struct rtattr *attr, const char *name)
 {
-	DBG("  attr %s (len %d) %s\n", name, (int) RTA_PAYLOAD(attr),
+	print("  attr %s (len %d) %s\n", name, (int) RTA_PAYLOAD(attr),
 						(char *) RTA_DATA(attr));
 }
 
 static inline void print_byte(struct rtattr *attr, const char *name)
 {
-	DBG("  attr %s (len %d) 0x%02x\n", name, (int) RTA_PAYLOAD(attr),
+	print("  attr %s (len %d) 0x%02x\n", name, (int) RTA_PAYLOAD(attr),
 					*((unsigned char *) RTA_DATA(attr)));
 }
 
+static inline void print_integer(struct rtattr *attr, const char *name)
+{
+	print("  attr %s (len %d) %d\n", name, (int) RTA_PAYLOAD(attr),
+						*((int *) RTA_DATA(attr)));
+}
+
 static inline void print_attr(struct rtattr *attr, const char *name)
 {
-	if (name)
-		DBG("  attr %s (len %d)\n", name, (int) RTA_PAYLOAD(attr));
+	int len = (int) RTA_PAYLOAD(attr);
+
+	if (name && len > 0)
+		print("  attr %s (len %d)\n", name, len);
 	else
-		DBG("  attr %d (len %d)\n",
-				attr->rta_type, (int) RTA_PAYLOAD(attr));
+		print("  attr %d (len %d)\n", attr->rta_type, len);
 }
 
 static void rtnl_link(struct nlmsghdr *hdr)
 {
-#if 0
 	struct ifinfomsg *msg;
 	struct rtattr *attr;
 	int bytes;
@@ -286,22 +597,22 @@ static void rtnl_link(struct nlmsghdr *hdr)
 	msg = (struct ifinfomsg *) NLMSG_DATA(hdr);
 	bytes = IFLA_PAYLOAD(hdr);
 
-	DBG("ifi_index %d ifi_flags 0x%04x", msg->ifi_index, msg->ifi_flags);
+	print("ifi_index %d ifi_flags 0x%04x", msg->ifi_index, msg->ifi_flags);
 
 	for (attr = IFLA_RTA(msg); RTA_OK(attr, bytes);
 					attr = RTA_NEXT(attr, bytes)) {
 		switch (attr->rta_type) {
 		case IFLA_ADDRESS:
-			print_attr(attr, "address");
+			print_ether(attr, "address");
 			break;
 		case IFLA_BROADCAST:
-			print_attr(attr, "broadcast");
+			print_ether(attr, "broadcast");
 			break;
 		case IFLA_IFNAME:
-			print_char(attr, "ifname");
+			print_string(attr, "ifname");
 			break;
 		case IFLA_MTU:
-			print_attr(attr, "mtu");
+			print_integer(attr, "mtu");
 			break;
 		case IFLA_LINK:
 			print_attr(attr, "link");
@@ -328,7 +639,7 @@ static void rtnl_link(struct nlmsghdr *hdr)
 			print_attr(attr, "protinfo");
 			break;
 		case IFLA_TXQLEN:
-			print_attr(attr, "txqlen");
+			print_integer(attr, "txqlen");
 			break;
 		case IFLA_MAP:
 			print_attr(attr, "map");
@@ -347,39 +658,26 @@ static void rtnl_link(struct nlmsghdr *hdr)
 			break;
 		}
 	}
-#endif
 }
 
 static void rtnl_newlink(struct nlmsghdr *hdr)
 {
-	struct ifinfomsg *msg;
-
-	msg = (struct ifinfomsg *) NLMSG_DATA(hdr);
-
-	DBG("ifi_type %d ifi_index %d ifi_flags 0x%04x ifi_change 0x%04x",
-					msg->ifi_type, msg->ifi_index,
-					msg->ifi_flags, msg->ifi_change);
-
-	process_newlink(msg->ifi_type, msg->ifi_index,
-					msg->ifi_flags, msg->ifi_change);
+	struct ifinfomsg *msg = (struct ifinfomsg *) NLMSG_DATA(hdr);
 
 	rtnl_link(hdr);
+
+	process_newlink(msg->ifi_type, msg->ifi_index, msg->ifi_flags,
+				msg->ifi_change, msg, IFA_PAYLOAD(hdr));
 }
 
 static void rtnl_dellink(struct nlmsghdr *hdr)
 {
-	struct ifinfomsg *msg;
-
-	msg = (struct ifinfomsg *) NLMSG_DATA(hdr);
-
-	DBG("ifi_type %d ifi_index %d ifi_flags 0x%04x ifi_change 0x%04x",
-					msg->ifi_type, msg->ifi_index,
-					msg->ifi_flags, msg->ifi_change);
-
-	process_dellink(msg->ifi_type, msg->ifi_index,
-					msg->ifi_flags, msg->ifi_change);
+	struct ifinfomsg *msg = (struct ifinfomsg *) NLMSG_DATA(hdr);
 
 	rtnl_link(hdr);
+
+	process_dellink(msg->ifi_type, msg->ifi_index, msg->ifi_flags,
+				msg->ifi_change, msg, IFA_PAYLOAD(hdr));
 }
 
 static void rtnl_addr(struct nlmsghdr *hdr)
@@ -391,7 +689,7 @@ static void rtnl_addr(struct nlmsghdr *hdr)
 	msg = (struct ifaddrmsg *) NLMSG_DATA(hdr);
 	bytes = IFA_PAYLOAD(hdr);
 
-	DBG("ifa_family %d ifa_index %d", msg->ifa_family, msg->ifa_index);
+	print("ifa_family %d ifa_index %d", msg->ifa_family, msg->ifa_index);
 
 	for (attr = IFA_RTA(msg); RTA_OK(attr, bytes);
 					attr = RTA_NEXT(attr, bytes)) {
@@ -403,7 +701,7 @@ static void rtnl_addr(struct nlmsghdr *hdr)
 			print_inet(attr, "local", msg->ifa_family);
 			break;
 		case IFA_LABEL:
-			print_char(attr, "label");
+			print_string(attr, "label");
 			break;
 		case IFA_BROADCAST:
 			print_inet(attr, "broadcast", msg->ifa_family);
@@ -424,9 +722,28 @@ static void rtnl_addr(struct nlmsghdr *hdr)
 	}
 }
 
+static void rtnl_newaddr(struct nlmsghdr *hdr)
+{
+	struct ifaddrmsg *msg = (struct ifaddrmsg *) NLMSG_DATA(hdr);
+
+	rtnl_addr(hdr);
+
+	process_newaddr(msg->ifa_family, msg->ifa_prefixlen, msg->ifa_index,
+						msg, IFA_PAYLOAD(hdr));
+}
+
+static void rtnl_deladdr(struct nlmsghdr *hdr)
+{
+	struct ifaddrmsg *msg = (struct ifaddrmsg *) NLMSG_DATA(hdr);
+
+	rtnl_addr(hdr);
+
+	process_deladdr(msg->ifa_family, msg->ifa_prefixlen, msg->ifa_index,
+						msg, IFA_PAYLOAD(hdr));
+}
+
 static void rtnl_route(struct nlmsghdr *hdr)
 {
-#if 0
 	struct rtmsg *msg;
 	struct rtattr *attr;
 	int bytes;
@@ -434,7 +751,10 @@ static void rtnl_route(struct nlmsghdr *hdr)
 	msg = (struct rtmsg *) NLMSG_DATA(hdr);
 	bytes = RTM_PAYLOAD(hdr);
 
-	DBG("rtm_family %d rtm_flags 0x%04x", msg->rtm_family, msg->rtm_flags);
+	print("rtm_family %d rtm_table %d rtm_protocol %d",
+			msg->rtm_family, msg->rtm_table, msg->rtm_protocol);
+	print("rtm_scope %d rtm_type %d rtm_flags 0x%04x",
+				msg->rtm_scope, msg->rtm_type, msg->rtm_flags);
 
 	for (attr = RTM_RTA(msg); RTA_OK(attr, bytes);
 					attr = RTA_NEXT(attr, bytes)) {
@@ -446,10 +766,10 @@ static void rtnl_route(struct nlmsghdr *hdr)
 			print_inet(attr, "src", msg->rtm_family);
 			break;
 		case RTA_IIF:
-			print_char(attr, "iif");
+			print_string(attr, "iif");
 			break;
 		case RTA_OIF:
-			print_attr(attr, "oif");
+			print_integer(attr, "oif");
 			break;
 		case RTA_GATEWAY:
 			print_inet(attr, "gateway", msg->rtm_family);
@@ -464,48 +784,51 @@ static void rtnl_route(struct nlmsghdr *hdr)
 			print_attr(attr, "metrics");
 			break;
 		case RTA_TABLE:
-			print_attr(attr, "table");
+			print_integer(attr, "table");
 			break;
 		default:
 			print_attr(attr, NULL);
 			break;
 		}
 	}
-#endif
+}
+
+static connman_bool_t is_route_rtmsg(struct rtmsg *msg)
+{
+
+	if (msg->rtm_table != RT_TABLE_MAIN)
+		return FALSE;
+
+	if (msg->rtm_protocol != RTPROT_BOOT &&
+			msg->rtm_protocol != RTPROT_KERNEL)
+		return FALSE;
+
+	if (msg->rtm_type != RTN_UNICAST)
+		return FALSE;
+
+	return TRUE;
 }
 
 static void rtnl_newroute(struct nlmsghdr *hdr)
 {
-	struct rtmsg *msg;
-
-	msg = (struct rtmsg *) NLMSG_DATA(hdr);
-
-	if (msg->rtm_type == RTN_UNICAST && msg->rtm_table == RT_TABLE_MAIN &&
-					msg->rtm_scope == RT_SCOPE_UNIVERSE) {
-		DBG("rtm_table %d rtm_scope %d rtm_type %d rtm_flags 0x%04x",
-					msg->rtm_table, msg->rtm_scope,
-					msg->rtm_type, msg->rtm_flags);
-		process_newgateway(msg, RTM_PAYLOAD(hdr));
-	}
+	struct rtmsg *msg = (struct rtmsg *) NLMSG_DATA(hdr);
 
 	rtnl_route(hdr);
+
+	if (is_route_rtmsg(msg))
+		process_newroute(msg->rtm_family, msg->rtm_scope,
+						msg, RTM_PAYLOAD(hdr));
 }
 
 static void rtnl_delroute(struct nlmsghdr *hdr)
 {
-	struct rtmsg *msg;
-
-	msg = (struct rtmsg *) NLMSG_DATA(hdr);
-
-	if (msg->rtm_type == RTN_UNICAST && msg->rtm_table == RT_TABLE_MAIN &&
-					msg->rtm_scope == RT_SCOPE_UNIVERSE) {
-		DBG("rtm_table %d rtm_scope %d rtm_type %d rtm_flags 0x%04x",
-					msg->rtm_table, msg->rtm_scope,
-					msg->rtm_type, msg->rtm_flags);
-		process_delgateway(msg, RTM_PAYLOAD(hdr));
-	}
+	struct rtmsg *msg = (struct rtmsg *) NLMSG_DATA(hdr);
 
 	rtnl_route(hdr);
+
+	if (is_route_rtmsg(msg))
+		process_delroute(msg->rtm_family, msg->rtm_scope,
+						msg, RTM_PAYLOAD(hdr));
 }
 
 static const char *type2string(uint16_t type)
@@ -648,8 +971,10 @@ static void rtnl_message(void *buf, size_t len)
 			rtnl_dellink(hdr);
 			break;
 		case RTM_NEWADDR:
+			rtnl_newaddr(hdr);
+			break;
 		case RTM_DELADDR:
-			rtnl_addr(hdr);
+			rtnl_deladdr(hdr);
 			break;
 		case RTM_NEWROUTE:
 			rtnl_newroute(hdr);
@@ -688,7 +1013,7 @@ static gboolean netlink_event(GIOChannel *chan,
 	return TRUE;
 }
 
-int connman_rtnl_send_getlink(void)
+static int send_getlink(void)
 {
 	struct rtnl_request *req;
 
@@ -708,7 +1033,27 @@ int connman_rtnl_send_getlink(void)
 	return queue_request(req);
 }
 
-int connman_rtnl_send_getroute(void)
+static int send_getaddr(void)
+{
+	struct rtnl_request *req;
+
+	DBG("");
+
+	req = g_try_malloc0(RTNL_REQUEST_SIZE);
+	if (req == NULL)
+		return -ENOMEM;
+
+	req->hdr.nlmsg_len = RTNL_REQUEST_SIZE;
+	req->hdr.nlmsg_type = RTM_GETADDR;
+	req->hdr.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
+	req->hdr.nlmsg_pid = 0;
+	req->hdr.nlmsg_seq = request_seq++;
+	req->msg.rtgen_family = AF_INET;
+
+	return queue_request(req);
+}
+
+static int send_getroute(void)
 {
 	struct rtnl_request *req;
 
@@ -728,6 +1073,11 @@ int connman_rtnl_send_getroute(void)
 	return queue_request(req);
 }
 
+int __connman_rtnl_request_update(void)
+{
+	return send_getlink();
+}
+
 int __connman_rtnl_init(void)
 {
 	struct sockaddr_nl addr;
@@ -741,9 +1091,7 @@ int __connman_rtnl_init(void)
 
 	memset(&addr, 0, sizeof(addr));
 	addr.nl_family = AF_NETLINK;
-	addr.nl_groups = RTMGRP_LINK | RTMGRP_IPV4_ROUTE;
-	//addr.nl_groups = RTMGRP_LINK | RTMGRP_IPV4_IFADDR;
-	//addr.nl_groups = RTMGRP_LINK | RTMGRP_IPV4_IFADDR | RTMGRP_IPV4_ROUTE;
+	addr.nl_groups = RTMGRP_LINK | RTMGRP_IPV4_IFADDR | RTMGRP_IPV4_ROUTE;
 
 	if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
 		close(sk);
@@ -759,6 +1107,15 @@ int __connman_rtnl_init(void)
 	return 0;
 }
 
+void __connman_rtnl_start(void)
+{
+	DBG("");
+
+	send_getlink();
+	send_getaddr();
+	send_getroute();
+}
+
 void __connman_rtnl_cleanup(void)
 {
 	GSList *list;
